"""Convert data from ColTopo (CT) to easier formats for calculation.

Segment :: (seg) a line in CT, typically depicting a route or trail.
Leg :: A portion of a segment ending with a labeled waypoint.
   But ONLY if waypoint is close to segment line.
Day :: A sequence of Segments intended to be traveled in a day. (but
   we might have to change our mind!)

Break :: (e.g. for rest) is specified in Marker/Comments with e.g. "20 min"
  It will show as Leg=Break in Travel Plan and be included in accum trav time.

Add some useful columns: segment, break_minutes
Add some useful Waypoints:
  camp|car :: waypoint at end of a day
"""
# See also: backpack-planning-artifacts.org
import csv
from pathlib import Path
import pandas as pd
import datetime as dt

def hr_min2minutes(hrmin):
    "Convert '5 hr 17 min' or '10 min' to int minutes"

    iterators = [iter(hrmin.split())] * 2
    dur = {k:int(v) for v,k in zip(*iterators, strict=True)}
    minutes = dur.get('hr',0)*60 + dur.get('min',0)
    return minutes

normal_rec = {
    'Leg': '0',
    'Gain': "+0'",
    'Loss': "-0'",
    'Distance': '0 ft',
    'Leg Time': '0 min',
    'Total Time': '0 min',
    }

# To pprint csv in linux:
#   echo "WP,LEG,LAT,LON,ELEV,GAIN,LOS,DIST,BEAR,LTIME,TTIME,NOTES" >hdr.csv
#   cat hdr.csv travelplan.GC-LCR-noraft.csv | column  -t -s,
# Extract from CalTopo Travel Plan CSV file (single segment)
def travelplan_multi_csv(csvfname='~/Downloads/travelplan.csv',
                         excelfname=None):
    """CSV file produced by:
    + click on line Bulk Ops
    + ctl-click on segments (labels) to include
      Click in the order they should be in the report.
    + Travel Plan
    + OPTIONAL: modify Travel mode and associated values for any Segement.
       This will affect the Leg and Total hike times.
    + Export to CSV.

    The resulting CSV has formatting issues interfer with calculations.
    This function tries to eliminate those issues.

    RETURN: pd.DataFrame
    """
    records = list()
    eod = False # End Of Day
    total_minutes = 0
    total_distance = 0
    prev = dict()
    new_rec = None
    trip_leg = 0
    elapsed_minutes = 0
    camp_cnt = 0
    day = travel_day = 1
    start_latlon = None

    header = ['Waypoint',    # the Label on a CT Marker
              'Leg',         # int; ID of leg within a segment (excludes Breaks)
              'Coordinates', # IGNORED; lat,lon
              'Elevation',   # (feet) of Waypoint
              'Gain',        # (feet) Accumulated Elevation Gain (aeg) over Leg
              'Loss',        # (feet) Accumulated Elevation Loss (ael) over Leg
              'Distance',    # (miles) Traveled distance over Leg
              'Bearing',     # IGNORED; nnnDegrees TN (True North)
              'Leg Time',    # (minutes) Traveled time over Leg
              'Total Time',  # (hrs:min) Accumulated travel time over Segment
              'Notes',
              ]

    with open(Path(csvfname).expanduser(), newline='') as csvfile:
        reader = csv.DictReader(csvfile, fieldnames=header)
        for rec in reader:
            # deal with special rows that to match the header
            if not any(rec.values()):  # empty row
                continue
            if not any(list(rec.values())[1:]):
                # If only a Waypoint value in row, we have new new Segment.
                seg_name = rec['Waypoint']
                continue
            if rec['Leg'] == 'Break':
                new_rec['break_minutes'] = hr_min2minutes(rec['Leg Time'])
                continue

            waypoint = rec['Waypoint']
            #!print(f"{rec['Waypoint']=} {camp_cnt=} {day=} {travel_day=}")

            match waypoint:
                case 'Start': # Start of segment (line for route or track)
                    eod = False
                    rec.update(normal_rec)
                    if start_latlon is None:
                        start_elevation = int(rec['Elevation'].strip("'"))
                        start_latlon = [float(n)
                                        for n in rec['Coordinates'].split(',')]
                case 'camp' | 'car': # "End" means of seg, this is of Day
                    eod = True
                    camp_cnt += 1
                    travel_day += 1
                case 'End': # End of segment (line for route or track)
                    # Replace "End" with "EOD" but only at end of day
                    # There is always a "total" WP for very segment, but
                    # not quite always an "End" WP!!? If so, tweak in CalTopo.
                    eod_rec = {
                        'day': day,
                        'eod': eod,
                        'segment': seg_name,
                        'waypoint': 'EOD',
                        'trip_leg': f"EOD{day}",
                        'elevation': int(prev['Elevation'].strip("'")),
                        'aeg': 0,
                        'ael': 0,
                        'distance': 0,
                        'duration': 0,
                        'break_minutes':  0,
                        'elapsed': elapsed_minutes,
                        'notes': '',
                    }
                    # Add a record after the last End of a day
                    if eod:
                        records.append(eod_rec)
                    continue
                case 'total': # we do our own total calculations
                    # There is always a "total" WP for very segment, but
                    # not quite always an "End" WP!!?
                    # The "total" row is anomolous and applies to SEGMENT
                    # so ignore.  We will calculate DAY totals ourself.
                    day = travel_day
                    continue
                case _: # Any other waypoint
                    pass

            minutes = hr_min2minutes(rec['Leg Time'])
            elapsed_minutes += minutes

            if new_rec:
                records.append(new_rec)

            if rec['Distance'].endswith(' mi'):
                distance = float(rec['Distance'].strip(" mi"))
            elif rec['Distance'].endswith(' ft'):
                distance = float(rec['Distance'].strip(" ft")) / 5280
            else:
                distance = 0.0
            lat, lon = [float(n) for n in rec['Coordinates'].split(',')]
            if rec['Gain'] == "":
                rec['Gain'] = "+0'"
            if rec['Loss'] == "":
                rec['Loss'] = "-0'"
            trip_leg += 1
            new_rec = {
                'day': day,
                'eod': eod,
                'segment': seg_name,
                'waypoint': rec['Waypoint'],
                'trip_leg': trip_leg,
                'elevation': int(rec['Elevation'].strip("'")),
                'aeg': int(rec['Gain'].strip("'")),
                'ael': int(rec['Loss'].strip("'")),
                'distance': distance,
                'duration': minutes,
                'break_minutes':  0,
                'elapsed': elapsed_minutes,
                'notes': rec['Notes'],
            }
            prev = rec
            # Add a record after the last End of a day
            if waypoint == 'End' and eod:
                records.append(eod_rec)
            # END: for rec in reader
        records.append(new_rec)
        #records.append(eod_rec)

        assert camp_cnt > 1, (
            f'For an overnight trip, you need at least one "camp" '
            f'and one "car" waypoint. (got {camp_cnt=})'
            )


        trip = dict(
            start_latlon=start_latlon, # decimal degrees: (lat,lon)
            start_elevation=start_elevation,  # feet
            elapsed_hours = elapsed_minutes/60,
            )


        df = pd.DataFrame(records)
        # Insert columns
        # Hours is "elapsed hours" which includes schedule breaks
        df['hours'] = (df['duration'] + df['break_minutes'])/ 60
        # pace excludes scheduled break_minutes.
        # Assume scheduled breaks are not "gotta rest now" kind of break.
        df['pace'] = (df['duration'] / (df['distance'] + 1e-6)).astype(int)
        col_types = {'elevation': int,
                     'aeg': int,
                     'ael': int,
                     'distance': float,
                     'duration': int,
                     'elapsed': int,  # total minutes (hike + break)
                     'break_minutes': int,
                     'hours': float,
                     'pace': int,
                     }
        df.astype(col_types)
        #! # Insert rows
        #! total_label = 'Total'
        #! total_row = pd.Series(df.sum(axis=0), name=total_label)
        #! df = pd.concat([df, pd.DataFrame(total_row).T], ignore_index=False)
        #!
        #! # Remove senseless values
        #! df.loc[total_label, 'day'] = pd.NA
        #! df.loc[total_label, 'segment'] = pd.NA
        #! df.loc[total_label, 'waypoint'] = pd.NA
        #! df.loc[total_label, 'leg'] = pd.NA
        #! df.loc[total_label, 'elevation'] = pd.NA

        # df columns:
        #   day, segment, waypoint, trip_leg, elevation, aeg, ael,
        #   distance, duration, break_minutes, elapsed, hours, pace

        # Write all data to Excel file
        if excelfname is None:
            excelfname = Path(csvfname).with_suffix('.xlsx')
        with pd.ExcelWriter(Path(excelfname).expanduser()) as writer:
            df.to_excel(writer, sheet_name='Waypoints', index=False)
            pd.DataFrame(trip).T.to_excel(writer,
                                          sheet_name='Meta', index=False)
        print(f'Wrote Excel version of dataframe to {excelfname}')
        return df, trip



import requests
import datetime as dt

# For much simpler retrieve of sunrise/set/twilights see:
# https://api.sunrise-sunset.org/json?lat=36.7201600&lng=-4.4203400
# {
#   "results": {
#     "sunrise": "5:08:31 AM",
#     "sunset": "7:19:35 PM",
#     "solar_noon": "12:14:03 PM",
#     "day_length": "14:11:04",
#     "civil_twilight_begin": "4:41:03 AM",
#     "civil_twilight_end": "7:47:02 PM",
#     "nautical_twilight_begin": "4:05:43 AM",
#     "nautical_twilight_end": "8:22:22 PM",
#     "astronomical_twilight_begin": "3:27:28 AM",
#     "astronomical_twilight_end": "9:00:38 PM"
#   },
#   "status": "OK",
#   "tzid": "UTC"
# }
#
# https://aa.usno.navy.mil/data/api
# https://aa.usno.navy.mil/api/rstt/oneday?date=2025-05-15&coords=36.08695,%20-111.86464&tz=-7
# Results with fields I don't care about removed
# {
#   "apiversion": "4.0.1",
#   "geometry": { "coordinates": [-111.86464,  36.08695 ],  },
#   "properties": {
#     "data": {
#       "closestphase": {
#         "day": 12, "month": 5, "phase": "Full Moon",
#         "time": "09:56",  "year": 2025
#       },
#       "curphase": "Waning Gibbous",
#       "day": 15,
#       "day_of_week": "Thursday",
#       "fracillum": "91%",
#       "isdst": false,
#       "label": null,
#       "month": 5,
#       "moondata": [
#         {"phen": "Set",  "time": "07:06"},
#         {"phen": "Rise", "time": "22:47"}
#       ],
#       "sundata": [
#         {"phen": "Begin Civil Twilight",  "time": "04:53"},
#         {"phen": "Rise", "time": "05:21"},
#         {"phen": "Set", "time": "19:27"},
#         {"phen": "End Civil Twilight", "time": "19:56"}
#       ],
#       "tz": -7.0,
#       "year": 2025
#     }
#   },
# }

def sun_from_lat_lon(lat: float,
                     lon: float,
                     date: int | str ='today',  # YYYYMMDD
                     tzid: str = 'America/Phoenix'
                     ):
    formatted: int = 0
    # See https://sunrise-sunset.org/api
    # Parameters: lat, lng, date, callback, formatted, tzid
    url = (f'https://api.sunrise-sunset.org/json?'
           f'date={date}&lat={lat}&lng={lon}'
           f'&formatted={formatted}&tzid={tzid}'
           )
    #! print(f'Using {url=}')
    response = requests.get(url)
    result = response.json()
    sdict = result['results']
    day_length = sdict['day_length']   # minutes
    del sdict['day_length']
    sundatetimes = {k:dt.datetime.fromisoformat(v) for k,v in sdict.items()}
    sundatetimes['day_length'] = day_length
    return sundatetimes

def sun_set_rise(lat: float,
                 lon: float,
                 date: str | None=None,  # YYYY-MM-DD. default=today
                 tzid: str = 'America/Phoenix',
                 formatted: int = 1  # 1 for simple local time for TZID; else ISO
                 ):
    """Intended for CAMP lat/lon. Get sunset for DATE and sunrise for NEXT DAY.
    """
    arrive_day = dt.date.today() if date is None else dt.date.fromisoformat(date)
    #! print(f'{lat=}, {lon=}, {date=}')
    #! print(f'{arrive_day=}')
    depart_day = arrive_day + dt.timedelta(days=1)
    sunset = sun_from_lat_lon(lat, lon,
                              date=str(arrive_day),
                              tzid=tzid)['sunset']
    twilight = sun_from_lat_lon(lat, lon,
                               date=depart_day,
                               tzid=tzid)['civil_twilight_begin'] # 6 degree
    sunrise = sun_from_lat_lon(lat, lon,
                               date=depart_day,
                               tzid=tzid)['sunrise']


    return  (sunset, twilight, sunrise)



import sqlite3

# For each Day: sunrise, sunset, camp_prep_hours
#   Calc camp_arrive_time: sunset - camp_prep_hours

# For each Segment:
# GIVEN: segment_name(key), distance, aeg, ael, (maxElev, minElev)
# ESTIMATE: pace,
# CALCULATE:  hours, minutes

{'GC Escalante':
 {'day 1':
  {'Tanner': {'distance': 7.2, 'aeg': 414, 'ael': -5079, 'est_pace': 60}},
  {'d-1b':   {'distance': 2.69,'aeg': 192, 'ael': -237, 'est_pace': 60}}
  },
 {'day 2':
  {'d2-a': {'distance': 8.85, 'aeg': 2276, 'ael': -2374, 'est_pace': 60}},
  #!{'75-ck': {'distance': 5.44+0.83, 'aeg': 1497+285, 'ael': -1525 + -64,
  #!           'est_pace': 60}},
  #!{'Papago wall':{'distance': .74, 'aeg': 205, 'ael': -169, 'est_pace': 60}},
  },
 {'day 3':
  {'d3': {'distance': 6., 'aeg': 4742, 'ael': -271, 'est_pace': 60}},
  {'TH to Car': {'distance': 1.27, 'aeg': 135, 'ael': -1, 'est_pace': 20}},
  }
 }


# Calculate time (minutes) to hike a segment
class HikeTime():
    def __init__(self, distance, aeg, ael):
        self.distance = distance  # miles
        self.aeg = aeg # Accumulated Elevated Gain (feet)
        self.ael = aeg # Accumulated Elevated Loss (feet)

    def miles_to_km(miles):
        return miles * 1.60934

    def feet_to_m(feet):
        return feet * 0.3047992424196

    @property
    def nasmiths_time(self):
        # hrs = dist/3 + ascent/2000
        minutes = self.distance*60/3 + aeg*60/2000 - ael*60/2000
        return minutes

    @property
    def book_time(self):
        return 30*self.distance + 30*aeg/1000

    @property
    def munter_walkup_time(self):
        # hours =  (DISTANCE [km] + (ELEVATION [m]/100)) / RATE
        # Rate = 2. Off-trail travel (Bushwhacking)
        # Rate = 4. Walking or skiing uphill
        # Rate = 6. Walking downhill
        # Rate = 10. Skiing downhill
        rate = 4
        hours = (miles_to_km(self.distance) + feet_to_m(self.aeg)/100)/rate
        return hours*60

    @property
    def munter_walkdown_time(self):
        rate = 6
        hours = (miles_to_km(self.distance) + feet_to_m(-1*self.aegl/100)/rate
        return hours*60

    @property
    def munter_offtrail_time(self):
        rate = 2
        hours = (miles_to_km(self.distance) + feet_to_m(self.aeg)/100)/rate
        return hours*60

    # END class HikeTime

def create_db(dbfile="hike_segments.db"):
    con = sqlite3.connect(dbfile)
    cur = con.cursor()
    cur.exeucte("CREATE TABLE segment(name, distance, aeg, ael)")
    return cur

# lat,lon used for sunrise/set.  e.g. of end of day (camp or car)
# https://sunrisesunset.io/api/
# https://api.sunrisesunset.io/json?lat=38.907192&lng=-77.036873&timezone=UTC&date_start=1990-05-01&date_end=1990-07-01
def ingest_seg(segment_name, distance, aeg, ael,
               pace: int | None = None,  # minutes/mile
               minElev=None,
               maxElev=None,
               ):
    record = dict(name=segment_name,
                  distance=distance,
                  aeg=aeg,
                  ael=ael)
    cur = create_db()
    cur.execute("INSERT INTO segment VALUES (:name, :distance, :aeg, :ael)",
                (record,))

# "camp" is the end of the last segment for a day.  It might be the car.
def ingest_day(camp_name: str,
               camp_lat: float, # decimal degrees
               camp_lon: float, # decimal degrees
               camp_prep_hours: int = 2,
               ):
    sunrise = None  # for lat/lon
    sunset  = None  # for lat/lon



"""Create timeline for hike/adventure that uses estimated pace to predict
time to reach each waypoint. Use given day start time or a start derived from sunrise local to hike. Warn if estimate camp arrival is too close to sunset."""

import csv
from pathlib import Path
import datetime as dt
from pprint import pp

import pandas as pd

from wagen import sun
import wagen.caltopo_io as cti


# given start time for each day, calc end time and hours from sunset!!!
# Get elevation (meters) from lat,lon
# https://api.open-elevation.com/api/v1/lookup?locations=41.161758,-8.583933
def timeline(trip_df, trip,
             start_time: dt.datetime | None = None, # sunrise today
             tzid: str = 'America/Phoenix',
             morning_daylight: float=1.5, # reqd hours @ camp after twilight
             evening_daylight: float=2.0, # reqd hours @ camp before sunset
             include_dt = False,
             rename = False,
             ):
    """Assume hiking will start the same time every day.
    Consider adding "BEGIN" waypoint for day specific start time!!!
    """
    #!print(f'timeline: {trip=} {start_time=} {tzid=} {morning_daylight=} {evening_daylight=} {include_dt=}')

    trip_meta = trip.copy()
    df = trip_df.copy(deep=True)
    day_minutes = 0
    prev_done_time = None
    eod = False

    lat, lon = trip['start_latlon']
    sunset, twilight, sunrise = sun.sun_set_rise(
        lat,
        lon,
        date=str(start_time)[:10].replace('-',''),
        tzid=tzid,
        formatted=0)

    # Modify start_time if does not contain both date and meaningful time.
    if start_time is None:
        start_time = dt.datetime.today()

    if start_time.hour == 0 and start_time.minute == 0:
        start_time = start_time.replace(hour=twilight.hour,
                                        minute=twilight.minute,
                                        tzinfo=twilight.tzinfo,
                                        second=0, microsecond=0)
        start_time += dt.timedelta(hours=morning_daylight)


    first_date = start_time.date()
    day_start_time = start_time.time()
    trip_meta = dict(
        first_date=first_date.isoformat(),
        day_start_time=day_start_time.strftime("%H:%M"),
        twilight=twilight.isoformat(),
        sunrise=sunrise.isoformat(),
        sunset=sunset.isoformat(),
        start_latlon=trip['start_latlon'],
        start_elevation=trip['start_elevation'],
        morning_daylight=morning_daylight,
        evening_daylight=evening_daylight,
        tzid=tzid,
        )

    for index, row in trip_df.iterrows():
        # Only do this on the first segment Start of the day
        if row['waypoint'] == 'Start' and eod:
            day_minutes = 0
            eod = False
        day_minutes += row.duration + row.break_minutes
        #!df.at[index, 'day_minutes'] = day_minutes
        row_dt = (dt.datetime.combine(first_date,
                                      day_start_time,
                                      sunrise.tzinfo)
                  + dt.timedelta(days=row['day']-1, minutes=day_minutes))
        #!!! df.at[index, 'deviation'] = row_dt.date()
        # see: https://docs.google.com/spreadsheets/d/1-DNLDEDKi286bAzVmbptdajS_mAE8OeM1oqFAemgrF4/edit?usp=sharing

        if (row_dt.time() > (sunset
                             - dt.timedelta(hours=evening_daylight)).time()):
            df.at[index, 'alert'] = "LATE to CAMP"
        else:
            df.at[index, 'alert'] = ''


        df.at[index, 'date'] = row_dt.date()
        done_time = row_dt.time()
        ldt = '' if prev_done_time == done_time else done_time.strftime("%H:%M")
        prev_done_time = done_time
        df.at[index, 'time'] = ldt
        df.at[index, 'day_hours'] = day_minutes/60.0
        df.at[index, 'dt'] = row_dt

        #!if row['waypoint'] in ['camp','car']: # 'EOD':
        if row['waypoint'] in ['camp','car']:
            eod = True
    # END for trip_df.iterrows

    final_column_order = ['date','time','alert', 'day', 'waypoint',
                          'distance', 'aeg','ael','day_hours',
                          'duration', 'break_minutes', 'elapsed',
                          'elevation', 'hours', 'pace',
                          # 'segment', 'trip_leg',
                          ]
    if include_dt:
        final_column_order.append('dt')

    df = df[final_column_order]

    #df.drop(['segment','dt', 'day'], axis="columns", inplace=True)
    if rename:
        renames = {c: c.title() for c in df.columns}
        renames.update({
            'distance': 'Distance (miles)',
            'aeg': 'AEG (ft)',
            'ael': 'AEL (ft)',
            'day_hours': 'Day Travel Time (hrs)',
            'duration': 'Leg Travel Time (mins)',
            'break_minutes': 'Leg Break (mins)',
            'pace': 'Pace (mins/mi)',
            })
        df.rename(columns=renames, inplace=True)
        df.set_index(['Date', 'Time'], inplace=True)
    return df, trip_meta

def trip_timeline(csvfname, **kwargs):
    caltopo_df, trip = cti.travelplan_multi_csv(csvfname)
    df, meta = timeline(caltopo_df, trip, **kwargs)
    return df, trip, meta



import datetime as dt
from zoneinfo import ZoneInfo
import warnings

import astropy.coordinates
import pandas as pd
import pytz  # DEPRECATED in favor of zoneinfo
import zoneinfo
from astroplan import Observer
from astropy.time import Time
from astropy import units as u


# https://api.sunrise-sunset.org/json?lat=36.7201600&lng=-4.4203400
# {
#   "results": {
#     "sunrise": "5:08:31 AM",
#     "sunset": "7:19:35 PM",
#     "solar_noon": "12:14:03 PM",
#     "day_length": "14:11:04",
#     "civil_twilight_begin": "4:41:03 AM",
#     "civil_twilight_end": "7:47:02 PM",
#     "nautical_twilight_begin": "4:05:43 AM",
#     "nautical_twilight_end": "8:22:22 PM",
#     "astronomical_twilight_begin": "3:27:28 AM",
#     "astronomical_twilight_end": "9:00:38 PM"
#   },
#   "status": "OK",
#   "tzid": "UTC"
# }
#

geodetic_coords = {
    'Manua Kea': {
        'lat': 19.82333,
        'lon': -155.46833,
        'elevation': 4214,  # meters
        'tzkey': 'HST'
        },
    'tucson': {
        'lat': 32.30942,
        'lon': -110.82205,
        'elevation': 831,  # meters
        'tzkey': 'America/Phoenix',
        },
    }

# alm_gn = Almanac(**gtp.geodetic_coords['gemini_north'])
class Almanac():
    """Get almanac data for a night given a date.
    A dayobs is the date of the start of an observing night. Therefore
    for sunrise and morning twilight we get time on the date AFTER dayobs.
    For sunset and evening twilight we get time on date of dayobs.
    For moonrise/set we get the time nearest to the midnight after day_ob.
    All times are in UTC.  Convert afterwards if needed.
    For list of timeszones: pytz.all_timezones
    E.G. [s for s in pytz.all_timezones if 'HST' in s]
         [s for s in pytz.all_timezones if 'Chile' in s]

    To get list of site_names:
       astropy.coordinates.EarthLocation.get_site_names()
    Compare Gemini_South to
       https://www.timeanddate.com/astronomy/@-30.23946,-70.73954
    Compare Gemini_North to
       https://www.timeanddate.com/astronomy/@19.82438,-155.46916
    """

    def __init__(
        self,
        lat: float,
        lon: float,
        tzkey: str,
        elevation: float = 0.0, # meters
        date: int | None = None, # YYYYMMDD
        #!timezone: dt.timedelta,  # UTC offset
        # default to night ending TODAY
        #!dayobs: int | None = None,  # YYYYMMDD
        # site="gemini_north", # gemini_north | gemini_south
    ):
        self.tzinfo = ZoneInfo(tzkey)

        if date is None:
            date = int(str(dt.date.today()).replace('-',''))

        #! else:
        #!     dobs = str(dayobs)
        #! dome_tz = pytz.timezone(timezone)
        #! self.dome_noon = Time(
        #!     dome_tz.localize(dt.datetime.strptime(dobs + " 12:00", "%Y%m%d %H:%M"))
        #! )

        #!astro_day = dt.datetime.strptime(dobs, "%Y%m%d").date()
        #!astro_date = dt.datetime.strptime(dobs, "%Y%m%d")
        self.date = date
        # Midnight is ambigious wrt date. Use just before midnight for clarity.
        # We use this to get the NEAREST sunrise/sunset etc
        # (which are different local days)
        self.dt_pre_midnight = dt.datetime.combine(
            dt.datetime.strptime(str(date),'%Y%m%d').date(),
            dt.time(23,59,59),
            tzinfo=self.tzinfo)
        self.pre_midnight = Time(
            self.dt_pre_midnight,
            format="datetime",
            scale="utc"
        )


        with warnings.catch_warnings(action="ignore"):
            #!self.loc = astropy.coordinates.EarthLocation.of_site(site)
            #! self.loc = astropy.coordinates.EarthLocation.from_geodetic(
            #!     lon,
            #!     lat,
            #!     height=elevation
            #!     )
            #!self.observer = Observer(self.loc, timezone=timezone)
            self.observer = Observer(latitude=lat * u.deg,
                                     longitude=lon * u.deg,
                                     elevation=elevation * u.m,
                                     timezone=tzkey)
            #! self.astro_day = astro_day
            #! self.astro_midnight = self.observer.midnight(
            #!     Time(
            #!         astro_date,
            #!         format="datetime",
            #!         scale="utc",
            #!     ),
            #!     which="next",
            #! )
            self.get_moon()
            self.get_sun()


    def __repr__(self):
        return (
            f'{self.date=} '
            f'{self.observer} '
            #!f'{self.tzinfo=} '
            f'pre-midnight(UTC)='
            f'{str(self.pre_midnight.to_datetime(timezone=ZoneInfo("UTC")))} '
            )

    def get_moon(self):
        self.moon_rise_time = self.observer.moon_rise_time(
            self.pre_midnight, which="nearest"
        )
        self.moon_set_time = self.observer.moon_set_time(
            self.pre_midnight, which="nearest"
        )

        # Percent of moon lit
        self.moon_illum = self.observer.moon_illumination(self.pre_midnight)

    def get_sun(self):
        # ast(ronoimical) twilight: -18 degrees)
        obs = self.observer
        self.ast_twilight_morning = obs.twilight_morning_astronomical(
            self.pre_midnight, which="nearest"
        )
        self.ast_twilight_evening = obs.twilight_evening_astronomical(
            self.pre_midnight, which="nearest"
        )

        # nau(tical) twilight: -12 degrees)
        self.nau_twilight_morning = self.observer.twilight_morning_nautical(
            self.pre_midnight, which="nearest"
        )
        self.nau_twilight_evening = self.observer.twilight_evening_nautical(
            self.pre_midnight, which="nearest"
        )

        # civ(il) twilight: -6 degrees)
        self.civ_twilight_morning = self.observer.twilight_morning_civil(
            self.pre_midnight, which="nearest"
        )
        self.civ_twilight_evening = self.observer.twilight_evening_civil(
            self.pre_midnight, which="nearest"
        )

        self.sun_rise_time = self.observer.sun_rise_time(
            self.pre_midnight, which="nearest"
        )
        self.sun_set_time = self.observer.sun_set_time(
            self.pre_midnight, which="nearest"
        )

    @property
    def night_hours(self):
        day_delta = self.ast_twilight_morning - self.ast_twilight_evening
        return day_delta.to_value("hr")

    @property
    def night_time(self):
        """(nightStartDatetime, nightEndDatetime) using 18deg twilight"""
        # set,rise
        return (self.events()['sunset_18deg'], self.events()['sunrise_18deg'])

    def events(self, iso=False, tzkey=None):
        """Sun/Moon datetime in UTC. Use localize=True for Chile time."""
        events = dict(  # as astropy.Time
            moon_rise=self.moon_rise_time,
            moon_set=self.moon_set_time,
            sunrise_18deg=self.ast_twilight_morning,
            sunset_18deg=self.ast_twilight_evening,
            sunrise_12deg=self.nau_twilight_morning,
            sunset_12deg=self.nau_twilight_evening,
            sunrise_6deg=self.civ_twilight_morning,
            sunset_6deg=self.civ_twilight_evening,
            sunrise=self.sun_rise_time,
            sunset=self.sun_set_time,
        )

        if tzkey is None:
            events_dt = {
                k: self.observer.astropy_time_to_datetime(v)
                for k, v in events.items()
            }
        else:
            tz = ZoneInfo(tzkey)
            events_dt = {
                k: self.observer.astropy_time_to_datetime(v).astimezone(tz)
                for k, v in events.items()
            }

        if iso:
            return {
                k: v.isoformat(sep=" ", timespec="seconds")
                for k, v in events_dt.items()
            }
        else:
            return events_dt

    @property
    def dataframe(self):
        df = pd.DataFrame(
            [
                #!self.events(localize=True, iso=True),
                self.events(localize=False, iso=True),
            ]
        ).T
        df.columns = ["UTC"]
        df.index.name = "Event"
        return df.sort_values(by="UTC").reset_index().set_index("UTC")

    @property
    def as_dict(self):
        moon_rise_time = Time(self.moon_rise_time, precision=0).iso
        moon_set_time = Time(self.moon_set_time, precision=0).iso
        ast_twilight_morning = Time(self.ast_twilight_morning, precision=0).iso
        ast_twilight_evening = Time(self.ast_twilight_evening, precision=0).iso
        nau_twilight_morning = Time(self.nau_twilight_morning, precision=0).iso
        nau_twilight_evening = Time(self.nau_twilight_evening, precision=0).iso
        civ_twilight_morning = Time(self.civ_twilight_morning, precision=0).iso
        civ_twilight_evening = Time(self.civ_twilight_evening, precision=0).iso
        sun_rise_time = Time(self.sun_rise_time, precision=0).iso
        sun_set_time = Time(self.sun_set_time, precision=0).iso

        # Maybe it we should add a column of times for the Dome.
        # It would make it easier to do some kinds of sanity checks.
        # Then again, it might confuse the issues.
        # It depends on who will be looking this the most.
        # Observers in the Dome? People elsewhere?

        data_dict = {
            "": "UTC",
            "Moon Rise": moon_rise_time,
            "Moon Set": moon_set_time,
            "Moon Illumination": f"{self.moon_illum:.0%}",
            "Morning Astronomical Twilight": ast_twilight_morning,
            "Evening Astronomical Twilight": ast_twilight_evening,
            "Morning Nautical Twilight": nau_twilight_morning,
            "Evening Nautical Twilight": nau_twilight_evening,
            "Morning Civil Twilight": civ_twilight_morning,
            "Evening Civil Twilight": civ_twilight_evening,
            "Sun Rise": sun_rise_time,
            "Sun Set": sun_set_time,
        }
        help_dict = {
            "": "",
            "Moon Set": "",
            "Moon Rise": "",
            "Moon Illumination": "(% illuminated)",
            "Morning Astronomical Twilight": "(-18 degrees)",
            "Evening Astronomical Twilight": "(-18 degrees)",
            "Solar Midnight": "",
            "Morning Nautical Twilight": "(-12 degrees)",
            "Evening Nautical Twilight": "(-12 degrees)",
            "Morning Civil Twilight": "(-6 degrees)",
            "Evening Civil Twilight": "(-6 degrees)",
            "Sun Set": "",
            "Sun Rise": "",
        }
        return data_dict, help_dict

    # A time_log is a DF ordered and indexed with DatetimeIndex.
    def as_records(self):
        """Sun/Moon events indexed by UTC (ISO string truncated to seconds)"""
        return self.dataframe.reset_index().to_dict(orient="records")
